import logging
from ..client.abc.events import Events


class EventHandler(Events):

    def __init__(self, collection):
        super().__init__()
        self._collection = collection

    def on_reconnect(self):
        self._collection.on_reconnect()

    def on_node_status(self, _status):
        pass

    def on_warning(self, warn):
        logging.warning(f'{warn["warn_msg"]} ({warn["warn_code"]})')

    def on_watch_init(self, data):
        thing_dict = data['thing']
        thing_id = thing_dict.pop('#')
        thing = self._collection._things.get(thing_id)
        if thing is None:
            thing_dict['#'] = thing_id  # restore `thing_dict`
            logging.debug(
                f'Cannot init #{thing_id} since the thing is not registerd '
                f'for watching by collection `{self._collection._name}`')
            return

        thing.on_init(data['event'], thing_dict)
        if thing is self._collection:
            for procedure in data['procedures']:
                thing._set_procedure(procedure)
            for enum_info in data['enums']:
                thing._update_enum(enum_info)
            for type_info in data['types']:
                thing._update_type(type_info)

    def on_watch_update(self, data):
        thing_id = data['#']
        thing = self._collection._things.get(thing_id)
        if thing is None:
            logging.debug(
                f'Cannot update #{thing_id} since the thing is not registerd '
                f'for watching by collection `{self._collection._name}`')
            return

        thing.on_update(data['event'], data.pop('jobs'))

    def on_watch_delete(self, data):
        thing_id = data['#']
        thing = self._collection._things.get(thing_id)
        if thing is not None:
            # since weakref is used, the thing is probably already removed and
            # the code will not reach this point, unless there are references
            # left.
            thing.on_delete()

    def on_watch_stop(self, data):
        thing_id = data['#']
        thing = self._collection._things.get(thing_id)
        if thing is not None:
            thing.on_stop()




import re
import functools
import logging


class PropTypes:

    @staticmethod
    def any_(v, klass, collection):
        if isinstance(v, dict):
            if '#' in v:
                return PropTypes.thing_(v, klass, collection)
            if '%' in v:
                return collection._get_enum_member(*v['%'])
            if '$' in v:
                return PropTypes.set_(v, nested=functools.partial(
                    klass=klass,
                    collection=collection
                ))
            if '*' in v:
                pattern = v['*']
                flags = 0
                if pattern.endswith('i'):
                    flags |= re.IGNORECASE
                    pattern = pattern[:-2]
                else:
                    pattern = pattern[1:-1]
                return re.compile(pattern, flags)
            if '!':
                msg = v['error_msg']
                # TODO : Return correct exception
                return Exception(msg)
            logging.warning(f'unhandled dict: {v}')
        return v

    @staticmethod
    def thing_(v, klass, collection, watch=False):
        if not isinstance(v, dict):
            raise TypeError(f'expecting type `dict`, got `{type(v)}`')

        thing_id = v.pop('#')
        thing = collection._things.get(thing_id)

        if thing is None:
            thing = klass(collection, thing_id)
        else:
            watch = False

        type_id = v.pop('.', None)
        if type_id is None:
            thing.__dict__.update(v)
        else:
            fmap = collection._types.get(type_id)
            if fmap:
                thing.__dict__.update(zip(fmap, v['']))

        if watch and not thing:
            collection._add_pending(thing)

        return thing

    @staticmethod
    def enum_(v, collection):
        return collection._get_enum_member(*v['%'])

    @staticmethod
    def str_(v):
        if not isinstance(v, str):
            raise TypeError(f'expecting type `str`, got `{type(v)}`')
        return v

    @staticmethod
    def utf8_(v):
        if not isinstance(v, str):
            raise TypeError(f'expecting type `str`, got `{type(v)}`')
        return v

    @staticmethod
    def bytes_(v):
        if not isinstance(v, bytes):
            raise TypeError(f'expecting type `bytes`, got `{type(v)}`')
        return v

    @staticmethod
    def raw_(v, _types=(str, bytes)):
        if not isinstance(v, _types):
            raise TypeError(f'expecting type `bytes`, got `{type(v)}`')
        return v

    @staticmethod
    def bool_(v):
        if not isinstance(v, bool):
            raise TypeError(f'expecting type `bool`, got `{type(v)}`')
        return v

    @staticmethod
    def int_(v):
        if not isinstance(v, int):
            raise TypeError(f'expecting type `int`, got `{type(v)}`')
        return v

    @staticmethod
    def uint_(v):
        if not isinstance(v, int):
            raise TypeError(f'expecting type `int`, got `{type(v)}`')
        if v < 0:
            raise ValueError(f'expecting an integer value >= 0, got {v}')
        return v

    @staticmethod
    def pint_(v):
        if not isinstance(v, int):
            raise TypeError(f'expecting type `int`, got `{type(v)}`')
        if v <= 0:
            raise ValueError(f'expecting an integer value > 0, got {v}')
        return v

    @staticmethod
    def nint_(v):
        if not isinstance(v, int):
            raise TypeError(f'expecting type `int`, got `{type(v)}`')
        if v >= 0:
            raise ValueError(f'expecting an integer value < 0, got {v}')
        return v

    @staticmethod
    def float_(v):
        if not isinstance(v, float):
            raise TypeError(f'expecting type `float`, got `{type(v)}`')
        return v

    @staticmethod
    def number_(v):
        if not isinstance(v, (float, int)):
            raise TypeError(
                f'expecting type `int` or `float`, got `{type(v)}`')
        return v

    @staticmethod
    def array_(v, nested):
        if not isinstance(v, list):
            raise TypeError(f'expecting a `list`, got `{type(v)}`')
        return [nested(item) for item in v]

    @staticmethod
    def set_(v, nested):
        if not isinstance(v, dict):
            raise TypeError(f'expecting a `dict`, got `{type(v)}`')
        v = v['$']
        return {nested(item) for item in v}

    @staticmethod

    def nillable(v, func=None):
        return v if v is None else func(v)



import functools
from .proptypes import PropTypes
from .enum import Enum


class Prop:

    __slots__ = (
        'vconv',
        'nconv',
        'spec',
        'nillable',
        'model',
    )

    @staticmethod
    def get_conv(fname, is_nillable=False, **kwrags):
        func = getattr(PropTypes, f'{fname}_', None)
        if kwrags:
            func = functools.partial(func, **kwrags)
        if func:
            return functools.partial(PropTypes.nillable, func=func) \
                if is_nillable else func

    def __init__(self, spec, cb=None):
        self.vconv = (cb, )
        self.nconv = None
        self.spec = spec
        self.nillable = False
        self.model = None

    def unpack(self, collection):
        from .thing import Thing  # nopep8
        if callable(self.vconv):
            return

        cb_type, self.vconv = self.vconv[0], lambda: True
        spec = self.spec
        is_nillable = False

        if spec.endswith('?'):
            self.nillable = is_nillable = True
            spec = spec[:-1]

        assert spec, 'an empty specification is not allowed'

        if spec.startswith('['):
            assert spec.endswith(']'), 'missing `]` in specification'
            self.nconv = ('array', is_nillable)
            spec = spec[1:-1]
            is_nillable = spec.endswith('?')
            if is_nillable:
                spec = spec[:-1]

        elif spec.startswith('{'):
            assert spec.endswith('}'), 'missing `}` in specification'
            self.nconv = ('set', is_nillable)
            spec = spec[1:-1]
            is_nillable = False

        kwargs = {
            'klass': Thing,
            'collection': collection,
        }

        if not spec:
            nested = self.get_conv('any', **kwargs)
            self.vconv = self.get_conv(*self.nconv, nested=nested)
            self.nconv = nested
            return  # finished, this is an array or set

        if spec == 'Thing' and cb_type is None:
            spec, kwargs['watch'] = 'thing', True

        self.vconv = self.get_conv(
            spec, is_nillable, **kwargs
            if spec == 'any' or spec == 'thing' else {})

        if self.vconv is None:
            assert callable(cb_type), \
                f'type `{spec}` is not a build-in, a calleble is required'

            self.model = cb_type \
                if isinstance(cb_type, type) and issubclass(cb_type, Thing) \
                else cb_type()

            if isinstance(self.model, Enum):
                self.vconv = self.get_conv('enum', collection=collection)
                return

            name = self.model._type_name

            assert spec == name, \
                f'type `{name}` does not match the specification `{spec}`'

            assert hasattr(self.model, '_props'), \
                f'missing `_props`, type `{name}` ' \
                f'must be a subclass of `Thing`'

            kwargs = {
                'klass': self.model,
                'collection': collection,
                'watch': True,
            }
            self.vconv = self.get_conv('thing', is_nillable, **kwargs)

            for p in self.model._props.values():
                p.unpack(collection)

        if self.nconv is not None:
            vconf = self.vconv
            self.vconv = self.get_conv(*self.nconv, nested=vconf)
            self.nconv = vconf



def convert(arg):
    if isinstance(arg, dict):
        id = arg.get('#')
        return {
            k: convert(v) for k, v in arg.items()
        } if id is None else {'#': id}

    if isinstance(arg, set):
        return {convert(v) for v in arg}

    if isinstance(arg, (list, tuple)):
        return [convert(v) for v in arg]

    return arg



def _wrap(value, blobs):

    if value is None:
        return 'nil'
    if isinstance(value, bool):
        return 'true' if value else 'false'
    if isinstance(value, bytes):
        idx = len(blobs)
        name = f'blob{idx}'
        blobs[name] = value
        return name
    if isinstance(value, Member):
        return value.name()
    if isinstance(value, dict):
        thing_id = value.get('#')
        if thing_id is None:
            thing = ','.join(
                f'{k}:{_wrap(v, blobs)}'
                for k, v in value.items()
            )
            return f"{{{thing}}}"
        return f'#{thing_id}'
    if isinstance(value, (list, tuple)):
        return f"[{','.join(_wrap(v, blobs) for v in value)}]"

    return repr(value)


def fmt(val, blobs=None):
    if blobs is None:
        blobs = {}
    return _wrap(val, blobs)



import asyncio
import weakref
import functools
from typing import Iterable, Optional, Union, TextIO, Type, Any
from ..client import Client
from .eventhandler import EventHandler
from .thing import Thing
from .enum import Enum
from .prop import Prop


class Collection(Thing):

    __STRICT__ = True
    __AS_TYPE__ = False

    def __init__(self, name=None):
        self._things = weakref.WeakValueDictionary()
        self._name = \
            getattr(self, '__COLLECTION_NAME__', self.__class__.__name__) \
            if name is None else name
        self._scope = f'//{self._name}'
        self._pending = set()  # Thing ID's
        self._client = None  # use load, build or rebuild
        self._id = None
        self._types = {}  # mapping where keys are type_id
        self._enums = {}  # mapping where keys are enum_id
        self._conv_any = Prop.get_conv('any', klass=Thing, collection=self)
        self._conv_thing = Prop.get_conv('thing', klass=Thing, collection=self)

        for p in self._props.values():
            p.unpack(self)

    async def load(self, client: Client) -> None:
        assert self._client is None, 'This collection is already loaded'
        self._client = client
        id = await self._client.query('.id()', scope=self._scope)
        super().__init__(self, id)
        client.add_event_handler(EventHandler(self))
        await self._client.watch(id, scope=self._scope)

    async def build(
            self,
            client: Client,
            classes: Optional[Iterable[Type[Thing]]] = None,
            scripts: Optional[Iterable[Union[TextIO, str]]] = None,
            delete_if_exists: bool = False,
    ) -> None:
        """Build the collection in ThingsDB.

        This will create the collection in ThingsDB will default values for
        all collection properties.

        Args:
            client (Client):
                ThingsDB Client instance with an active, authenticated
                connection.
            classes (iterable, optional):
                Optional list of classes to create. This is only required when
                a class has no relation with the collection, otherwise the
                class will be created recursively while building the
                collection. Defaults to `None`.
            scripts (iterable, optional):
                Optional list of script which will be started after building
                the collection. They will be started in the same order as they
                are given. The iterable may contain File Objects in Text mode,
                or plain strings with ThingsDB code. Defaults to `None`.
            delete_if_exists (bool):
                When `True`, the collection will be removed if it exists. Be
                careful since all data in the collection will be removed! If
                this arguments is `False`, a `KeyError` will be raised when the
                collection exists. Defaults to `False`.
        """
        assert self._client is None, 'This collection is already loaded'
        if (await client.has_collection(self._name)):
            if delete_if_exists:
                await client.del_collection(self._name)
            else:
                raise KeyError(f'Collection `{self._name}` already exists')

        # create the collection, we are sure it does not exists
        await client.new_collection(self._name)

        # first create the types so circular dependencies may be handled
        await self._new_type(client, self)
        if classes:
            for model in classes:
                await model._new_type(client, self)

        # set the type definitions and
        await self._set_type(client, self)
        if classes:
            for model in classes:
                await model._set_type(client, self)

        if scripts is not None:
            for script in scripts:
                code = script if isinstance(script, str) else script.read()
                await client.query(
                    code=code,
                    scope=self._scope,
                    convert_vars=False)

    async def query(self, code: str, **kwargs: Any) -> Any:
        """Query using this collection as scope.

        This is the same as calling the `query(..)` method on the client with
        the scope='...' argument set to the collection scope. All keyword
        arguments will be parsed to the `Client().query(..)` method so look at
        that method for more information.
        """
        return await self._client.query(code, scope=self._scope, **kwargs)

    def on_reconnect(self):
        """Called from the `EventHandler`."""
        self._pending.update(self._things.keys())
        self._go_pending()

    def _add_pending(self, thing):
        self._pending.add(thing.id())

    def _go_pending(self):
        if not self._pending:
            return
        future = asyncio.ensure_future(
            self._client.watch(*self._pending, scope=self._scope),
            loop=self._client._loop
        )
        self._pending.clear()
        return future

    def _register(self, thing: Thing) -> None:
        self._things[thing._id] = thing

    def _set_procedure(self, data):
        name = data['name']
        setattr(self, name, functools.partial(
            self._client.run,
            name,
            scope=self._scope))

    def _update_type(self, data):
        self._types[data['type_id']] = tuple(k[0] for k in data['fields'])

    def _upd_type_add(self, data):
        self._types[data['type_id']] += data['name'],

    def _upd_type_del(self, data):
        type_id, name = data['type_id'], data['name']
        t = self._types[type_id]
        idx = t.index(name)
        t = list(t)
        try:
            t[idx] = t.pop()  # swap remove
        except IndexError:
            pass
        self._types[type_id] = tuple(t)

    def _upd_type_ren(self, data):
        type_id, name, to = data['type_id'], data['name'], data['to']
        t = self._types[type_id]
        idx = t.index(name)
        t = list(t)
        t[idx] = to
        self._types[type_id] = tuple(t)

    def _update_enum(self, data):
        Enum._update_enum(self._enums, data, convert=self._conv_any)

    def _upd_enum_add(self, data):
        Enum._upd_enum_add(self._enums, data, convert=self._conv_any)

    def _upd_enum_del(self, data):
        Enum._upd_enum_del(self._enums, data)

    def _upd_enum_mod(self, data):
        Enum._upd_enum_mod(self._enums, data, convert=self._conv_any)

    def _upd_enum_ren(self, data):
        Enum._upd_enum_ren(self._enums, data)

    def _get_enum_member(self, enum_id, idx):
        enum = self._enums[enum_id]
        return enum[idx]



import datetime
from typing import Union as U


class Buildin:

    async def collection_info(self, collection: U[int, str]) -> dict:
        return await self.query(
            f'collection_info({collection!r})',
            scope='@t')

    async def collections_info(self):
        return await self.query('collections_info()', scope='@t')

    async def counters(self, scope='@n'):
        return await self.query('counters()', scope=scope)

    async def del_collection(self, collection: U[int, str]):
        return await self.query(
            f'del_collection({collection!r})',
            scope='@t')

    async def del_expired(self):
        return await self.query('del_expired()', scope='@t')

    async def del_token(self, key: str):
        return await self.query(f'del_token({key!r})', scope='@t')

    async def del_user(self, name: str):
        return await self.query(f'del_user({name!r})', scope='@t')

    async def grant(self, target: U[int, str], user: str, mask: int):
        return await self.query(
            f'grant({target!r}, {user!r}, {mask})',
            scope='@t')

    async def has_collection(self, name: str):
        return await self.query(f'has_collection({name!r})', scope='@t')

    async def has_user(self, name: str):
        return await self.query(f'has_user({name!r})', scope='@t')

    async def new_collection(self, name: str):
        return await self.query(f'new_collection({name!r})', scope='@t')

    async def new_token(
            self,
            user: str,
            expiration_time: datetime.datetime = None,
            description: str = ''):

        if expiration_time is None:
            expiration_time = 'nil'
        else:
            expiration_time = int(datetime.datetime.timestamp(expiration_time))

        return await self.query(
            f'new_token({user!r}, {expiration_time}, {description!r})',
            scope='@t')

    async def node_info(self, scope='@n'):
        return await self.query('node_info()', scope=scope)

    async def nodes_info(self, scope='@n') -> list:
        return await self.query('nodes_info()', scope=scope)

    async def reset_counters(self, scope='@n') -> None:
        return await self.query('reset_counters()', scope=scope)

    async def revoke(self, target: U[int, str], user: str, mask: int):
        return await self.query(
            f'grant({target!r}, {user!r}, {mask})',
            scope='@t')

    async def set_log_level(self, log_level: str, scope='@n') -> None:
        assert log_level in ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL')
        return await self.query(f'set_log_level({log_level})', scope=scope)

    async def shutdown(self, scope='@n') -> None:
        return await self.query('shutdown()', scope=scope)

    async def user_info(self, user: str = None) -> dict:
        if user is None:
            return await self.query('user_info()', scope='@t')
        return await self.query(f'user_info({user!r})', scope='@t')

    async def users_info(self) -> list:
        return await self.query('users_info()', scope='@t')



from ..protocol import Proto
import abc
from typing import Any


class Events(abc.ABC):

    def __init__(self):
        self._evmap = {
            Proto.ON_NODE_STATUS: self.on_node_status,
            Proto.ON_WARN: self.on_warning,
            Proto.ON_WATCH_INI: self.on_watch_init,
            Proto.ON_WATCH_UPD: self.on_watch_update,
            Proto.ON_WATCH_DEL: self.on_watch_delete,
            Proto.ON_WATCH_STOP: self.on_watch_stop,
        }

    def __call__(self, tp: Proto, data: Any) -> None:
        self._evmap.get(tp)(data)

    @abc.abstractmethod
    def on_reconnect(self) -> None:
        """On re-connect
        Called after a re-concect is finished (including authentication)
        """
        pass

    @abc.abstractmethod
    def on_node_status(self, status: str) -> None:
        """On node status
        status: String containing a `new` node status.
                Optional values:
                    - OFFLINE
                    - CONNECTING
                    - BUILDING
                    - SHUTTING_DOWN
                    - SYNCHRONIZING
                    - AWAY
                    - AWAY_SOON
                    - READY
        """
        pass

    @abc.abstractmethod
    def on_warning(self, warn: dict) -> None:
        """On warning
        warn: a dictionary with `warn_msg` and `warn_code`. for example:

        {
            "warn_msg": "some warning message"
            "warn_code": 1
        }
        """
        pass

    @abc.abstractmethod
    def on_watch_init(self, data: dict) -> None:
        """On watch init.
        Initial data from a single thing. for example:

        {
            "#": 123,
            "name": "ThingsDB!",
            ...
        }
        """
        pass

    @abc.abstractmethod
    def on_watch_update(self, data: dict) -> None:
        """On watch update.
        Updates for a thing with ID (#). One event may contain more than one
        job. for example:

        {
            "#": 123,
            "jobs": [
                {
                    "set": {
                        "answer": 42
                    }
                }
            ]
        }
        """
        pass

    @abc.abstractmethod
    def on_watch_delete(self, data: dict) -> None:
        """On watch delete.
        The thing is removed from the collection (and garbage collected).
        for example:

        {
            "#": 123
        }
        """
        pass

    @abc.abstractmethod
    def on_watch_stop(self, data: dict) -> None:
        """On watch stop.
        The thing is not watched anymore due to either call to `unwatch()`, or
        by a unwatch request (REQ_UNWATCH). This event is *not* triggered when
        a connection to a node has been lost.

        for example:

        {
            "#": 123
        }
        """
        pass



import asyncio
import logging
from .prop import Prop


def checkevent(f):
    def wrapper(self, event_id, *args):
        if self._event_id > event_id:
            logging.warning(
                f'ignore event because the current event `{self._event_id}` '
                f'is greather than the received event `{event_id}`')
            return
        self._event_id = event_id
        f(self, event_id, *args)
        self._collection._go_pending()
    return wrapper


class ThingHash:
    def __init__(self, id):
        self._id = id

    def __hash__(self):
        return self._id

    def __eq__(self, other):
        return self._id == other._id


class Thing(ThingHash):
    # When __STRICT__ is set to `True`, only properties which are defined in
    # the model class are assigned to a `Thing` instance. If `False`, all
    # properties are set, not only the ones defined by the model class.
    __STRICT__ = False

    # When __SET_ANYWAY__ is set to `True`, values which do mot match the
    # specification will be assigned to a `Thing` instance anyway and only
    # a warning will be logged. If `False`, the properties will not be set.
    __SET_ANYWAY__ = False

    # When __AS_TYPE__ is set to `True`, this class will be created in
    # thingsdb as a Type when using the `build(..)` method. If `False`, no type
    # will be created. A Collection instance will have `False` as default.
    __AS_TYPE__ = True

    _props = dict()
    _type_name = None  # Only set when __AS_TYPE__ is True
    _visited = 0  # For build, 0=not visited, 1=new_type, 2=set_type, 3=build

    def __init__(self, collection, id: int):
        super().__init__(id)
        self._event_id = 0
        self._collection = collection
        collection._register(self)

    def __init_subclass__(cls):
        cls._props = {}
        items = {
            k: v for k, v in cls.__dict__.items() if not k.startswith('__')}
        for key, val in items.items():
            if isinstance(val, str):
                val = val,
            if isinstance(val, tuple):
                prop = cls._props[key] = Prop(*val)
                delattr(cls, key)
        if cls.__AS_TYPE__:
            cls._type_name = getattr(cls, '__TYPE_NAME__', cls.__name__)

    def __bool__(self):
        return bool(self._event_id)

    def __repr__(self):
        return f'#{self._id}'

    def id(self):
        return self._id

    def get_collection(self):
        return self._collection

    def get_client(self):
        return self._collection._client

    def watch(self):
        collection = self._collection
        return collection._client.watch(self._id, scope=collection._scope)

    def unwatch(self):
        collection = self._collection
        return collection._client.unwatch(self._id, scope=collection._scope)

    @checkevent
    def on_init(self, event, data):
        self._job_set(data)

    @checkevent
    def on_update(self, event, jobs):
        for job_dict in jobs:
            for name, job in job_dict.items():
                jobfun = self._UPDMAP.get(name)
                if jobfun is None:
                    raise TypeError(f'unknown job `{name}` for `{self}`')
                jobfun(self, job)

    def on_delete(self):
        self._collection._things.pop(self.id())

    def on_stop(self):
        logging.warning(f'stopped watching thing {self}')

    def _job_add(self, pair):
        cls = self.__class__
        (k, v), = pair.items()
        prop = cls._props.get(k)

        try:
            set_ = getattr(self, k)
        except AttributeError:
            if prop:
                logging.warning(
                    f'missing property `{k}` on `{self}` '
                    f'while the property is defined in the '
                    f'model class as `{prop.spec}`')
            return

        if not isinstance(set_, set):
            logging.warning(
                f'got a add job for property `{k}` on `{self}` '
                f'while the property is of type `{type(set_)}`')
            return

        convert = prop.nconv if prop else self._collection._conv_thing
        try:
            set_.update((convert(item) for item in v))
        except Exception as e:
            logging.warning(
                f'got a value for property `{k}` on `{self}` which '
                f'does not match `{prop.spec if prop else "thing"}` ({e})')

    def _job_del(self, k):
        prop = self.__class__._props.get(k)
        if prop:
            logging.warning(
                f'property `{k}` on `{self}` will be removed while it '
                f'is defined in the model class as `{prop.spec}`')
        try:
            delattr(self, k)
        except AttributeError:
            pass

    def _job_remove(self, pair):
        cls = self.__class__
        (k, v), = pair.items()

        try:
            set_ = getattr(self, k)
        except AttributeError:
            if prop:
                logging.warning(
                    f'missing property `{k}` on `{self}` '
                    f'while the property is defined in the '
                    f'model class as `{prop.spec}`')
            return

        if not isinstance(set_, set):
            logging.warning(
                f'got a remove job for property `{k}` on `{self}` '
                f'while the property is of type `{type(set_)}`')
            return

        set_.difference_update((ThingHash(id) for id in v))

    def _job_set(self, pairs):
        cls = self.__class__

        for k, v in pairs.items():
            prop = cls._props.get(k)
            if prop:
                convert = prop.vconv
            elif cls.__STRICT__:
                continue
            else:
                convert = self._collection._conv_any

            try:
                v = convert(v)
            except Exception as e:
                logging.warning(
                    f'got a value for property `{k}` on `{self}` which does '
                    f'not match `{prop.spec if prop else "any"}` ({repr(e)})')
                if not cls.__SET_ANYWAY__:
                    continue
            setattr(self, k, v)

        self._collection._go_pending()

    def _job_splice(self, pair):
        cls = self.__class__
        (k, v), = pair.items()
        prop = cls._props.get(k)

        try:
            arr = getattr(self, k)
        except AttributeError:
            if prop:
                logging.warning(
                    f'missing property `{k}` on `{self}` '
                    f'while the property is defined in the '
                    f'model class as `{prop.spec}`')
            return

        if not isinstance(arr, list):
            logging.warning(
                f'got a splice job for property `{k}` on `{self}` '
                f'while the property is of type `{type(arr)}`')
            return

        index, count, *items = v
        convert = prop.nconv if prop else self._collection._conv_any
        try:
            arr[index:index+count] = (convert(item) for item in items)
        except (TypeError, ValueError) as e:
            logging.warning(
                f'got a value for property `{k}` on `{self}` '
                f'which does not match `{prop.spec if prop else "any"}` ({e})')

    def _job_del_procedure(self, data):
        delattr(self._collection, data)

    def _job_del_type(self, data):
        # we can just as well keep the type so simply ignore this event
        pass

    def _job_mod_type_add(self, data):
        self._collection._upd_type_add(data)

    def _job_mod_type_del(self, data):
        self._collection._upd_type_del(data)

    def _job_mod_type_mod(self, data):
        # we do not care about the specification so simply ignore this event
        pass

    def _job_mod_enum_add(self, data):
        self._collection._upd_enum_add(data)

    def _job_mod_enum_del(self, data):
        self._collection._upd_enum_del(data)

    def _job_mod_enum_mod(self, data):
        self._collection._upd_enum_mod(data)

    def _job_mod_enum_ren(self, data):
        self._collection._upd_enum_ren(data)

    def _job_new_procedure(self, data):
        self._collection._set_procedure(data)

    def _job_new_type(self, data):
        data['fields'] = []
        self._collection._update_type(data)

    def _job_set_type(self, data):
        self._collection._update_type(data)

    _UPDMAP = {
        # Thing jobs
        'add': _job_add,
        'del': _job_del,
        'remove': _job_remove,
        'set': _job_set,
        'splice': _job_splice,

        # Collection jobs
        'del_procedure': _job_del_procedure,
        'del_type': _job_del_type,
        'mod_type_add': _job_mod_type_add,
        'mod_type_del': _job_mod_type_del,
        'mod_type_mod': _job_mod_type_mod,
        'mod_enum_add': _job_mod_enum_add,
        'mod_enum_del': _job_mod_enum_del,
        'mod_enum_mod': _job_mod_enum_mod,
        'mod_enum_ren': _job_mod_enum_ren,
        'new_procedure': _job_new_procedure,
        'new_type': _job_new_type,
        'set_type': _job_set_type,
    }

    @classmethod
    async def _new_type(cls, client, collection):
        if cls._visited > 0:
            return
        cls._visited += 1

        for prop in cls._props.values():
            if prop.model:
                await prop.model._new_type(client, collection)

        if not cls._type_name:
            return

        await client.query(f'''
            new_type('{cls._type_name}');
        ''', scope=collection._scope)

    @classmethod
    async def _set_type(cls, client, collection):
        if cls._visited > 1:
            return
        cls._visited += 1

        for prop in cls._props.values():
            if prop.model:
                await prop.model._set_type(client, collection)

        if not cls._type_name:
            return

        await client.query(f'''
            set_type('{cls._type_name}', {{
                {', '.join(f'{k}: "{p.spec}"' for k, p in cls._props.items())}
            }});
        ''', scope=collection._scope)


class ThingStrict(Thing):

    __STRICT__ = True



import struct
import msgpack


class Package(object):

    __slots__ = ('pid', 'length', 'total', 'tp', 'checkbit', 'data')

    st_package = struct.Struct('<IHBB')

    def __init__(self, barray: bytearray) -> None:
        self.length, self.pid, self.tp, self.checkbit = \
            self.__class__.st_package.unpack_from(barray, offset=0)
        self.total = self.__class__.st_package.size + self.length
        self.data = None

    def extract_data_from(self, barray: bytearray) -> None:
        try:
            self.data = msgpack.unpackb(
                barray[self.__class__.st_package.size:self.total],
                raw=False) \
                if self.length else None
        finally:
            del barray[:self.total]

    def __repr__(self) -> str:
        return '<id: {0.pid} size: {0.length} tp: {0.tp}>'.format(self)



from .member import Member

_enums_lookup = {}  # enums lookup by name

class Enum:

    _visited = 0  # For build, 0=not visited, 1=new_type, 2=set_type, 3=build

    def __init_subclass__(cls, **kwargs):
        cls._name = getattr(cls, '__NAME__', cls.__name__)
        cls._id = None

        # upgrade attributes to member instances
        for k, v in cls.__dict__.items():
            if k.startswith('_'):
                continue
            setattr(cls, k, Member(cls._name, k, v))

        # register for lookup by name
        _enums_lookup[cls._name] = cls

    @staticmethod
    def _update_enum(enums, data, convert):
        name = data['name']
        members = [Member(name, k, convert(v)) for k, v in data['members']]

        enum = _enums_lookup.get(name)
        if enum is not None:
            enum._id = data['enum_id']
            for member in members:
                setattr(enum, member.name, member)

        enums[data['enum_id']] = members

    @staticmethod
    def _upd_enum_add(enums, data, convert):
        members = enums[data['enum_id']]
        name = members[0]._enum_name
        member = Member(name, data['name'], convert(data['value']))
        members.append(member)

        enum = _enums_lookup.get(name)
        if enum is not None:
            setattr(enum, member.name, member)

    @staticmethod
    def _upd_enum_del(enums, data):
        members = enums[data['enum_id']]
        name = members[0]._enum_name

        enum = _enums_lookup.get(name)
        if enum is not None:
            member = members[data['index']]
            delattr(enum, member.name)

        try:
            # swap remove the index
            members[data['index']] = members.pop()
        except IndexError:
            pass

    @staticmethod
    def _upd_enum_mod(enums, data, convert):
        members = enums[data['enum_id']]
        name = members[0]._enum_name
        member = members[data['index']]
        member._value = convert(data['value'])

    @staticmethod
    def _upd_enum_ren(enums, data):
        members = enums[data['enum_id']]
        name = members[0]._enum_name
        member = members[data['index']]

        enum = _enums_lookup.get(name)
        if enum is not None:
            delattr(enum, member.name)
            setattr(enum, data['name'], member)

        member._name = data['name']

    @classmethod
    async def _new_type(cls, client, collection):
        if cls._visited > 0:
            return
        cls._visited += 1
        members = (m for m in cls.__dict__.values() if isinstance(m, Member))
        query = f'''
            set_enum('{cls._name}', {{
                {', '.join(f'{m.name}: {m.value!r}' for m in members)}
            }});
        '''
        await client.query(query, scope=collection._scope)

    @classmethod
    async def _set_type(cls, client, collection):
        pass

    @classmethod
    def id(cls):
        return cls._id

    @classmethod
    def name(cls):
        return cls._name


class Member:
    def __init__(self, enum_name, name, value):
        self._name = name
        self._value = value
        self._enum_name = enum_name

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        return self._value

    def __repr__(self):
        return f'{self._enum_name}{{{self._name}}}'



import asyncio
import logging
import random
import ssl
from ssl import SSLContext, PROTOCOL_TLS
from typing import Optional, Union, Any
from deprecation import deprecated
from .buildin import Buildin
from .protocol import Proto
from .protocol import Protocol
from .abc.events import Events
from ..util.convert import convert
from ..exceptions import ForbiddenError


_WATCH_MISSING = \
    'auto reconnect cannot act on node changes since `WATCH` privileges on ' \
    'the `@node` scope are missing'


class Client(Buildin):

    MAX_RECONNECT_WAIT_TIME = 120
    MAX_RECONNECT_TIMEOUT = 10

    def __init__(
            self,
            auto_reconnect: bool = True,
            ssl: Optional[Union[bool, ssl.SSLContext]] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        """Initialize a ThingsDB client.

        Args:
            auto_reconnect (bool, optional):
                When set to `True`, the client will automatically
                reconnect when a connection is lost. If set to `False` and the
                connection gets lost, one may call the `reconnect()` method to
                make a new connection. The auto-reconnect option can act on
                node changes and does so automatically if the connected user
                has the required `WATCH` privileges on the `@node` scope.
                Defaults to `True`.
            ssl (SSLContext or bool, optional):
                Accepts an ssl.SSLContext for creating a secure connection
                using SSL/TLS. This argument may simply be set to `True` in
                which case a context using `ssl.PROTOCOL_TLS` is created.
                Defaults to None.
            loop (AbstractEventLoop, optional):
                Can be used to run the client on a specific event loop.
                If this argument is not used, the default event loop will be
                used. Defaults to `None`.
        """

        self._loop = loop if loop else asyncio.get_event_loop()
        self._auth = None
        self._pool = None
        self._protocol = None
        self._pid = 0
        self._reconnect = auto_reconnect
        self._scope = '@t'  # default to thingsdb scope
        self._pool_idx = 0
        self._reconnecting = False
        if ssl is True:
            self._ssl = SSLContext(PROTOCOL_TLS)
        elif ssl is False:
            self._ssl = None
        else:
            self._ssl = ssl
        self._event_handlers = []

    def add_event_handler(self, event_handler: Events) -> None:
        """Add an event handler.

        Event handlers will called in the order they are added.

        Args:
            event_handler (Events):
                An instance of Events (see thingsdb.client.abc.events).
        """
        self._event_handlers.append(event_handler)

    def get_event_loop(self) -> asyncio.AbstractEventLoop:
        """Can be used to get the event loop.

        Returns:
            AbstractEventLoop: The event loop used by the client.
        """
        return self._loop

    def is_connected(self) -> bool:
        """Can be used to check if the client is connected.

        Returns:
            bool: `True` when the client is connected else `False`.
        """
        return bool(self._protocol and self._protocol.transport)

    def set_default_scope(self, scope: str) -> None:
        """Set the default scope.

        Can be used to change the default scope which is initially set to `@t`.

        Args:
            scope (str):
                Set the default scope. A scope may start with either the `/`
                character, or `@`. Examples: "//stuff", "@:stuff", "/node"
        """
        assert scope.startswith('@') or scope.startswith('/')
        self._scope = scope

    def get_default_scope(self) -> str:
        """Get the default scope.

        The default scope may be changed with the `set_default_scope()` method.

        Returns:
            str:
                The default scope which is used by the client when no specific
                scope is specified.
        """
        return self._scope

    def close(self) -> None:
        """Close the ThingsDB connection.

        This method will return immediately so the connection may not be
        closed yet after a call to `close()`. Use the `wait_closed()` method
        after calling this method if this is required.
        """
        if self._protocol and self._protocol.transport:
            self._reconnect = False
            self._protocol.transport.close()

    def connection_info(self) -> str:
        """Returns the current connection info as a string.

        Even with a connection pool, the client has still one active node
        connection at the time, and info for this active connection will be
        returned.

        example: "node0.local:9200"
        """
        if not self.is_connected():
            return 'disconnected'
        socket = self._protocol.transport.get_extra_info('socket', None)
        if socket is None:
            return 'unknown_addr'
        addr, port = socket.getpeername()
        return f'{addr}:{port}'

    def connect_pool(
            self,
            pool: list,
            *auth: Union[str, tuple]
    ) -> asyncio.Future:
        """Connect using a connection pool.

        When using a connection pool, the client will randomly choose a node
        to connect to. When a node is going down, it will inform the client
        so it will automatically re-connect to another node. Connections will
        automatically authenticate so the connection pool requires credentials
        to perform the authentication.

        Examples:
            >>> await connect_pool([
                'node01.local',             # address as string
                'node02.local',             # port will default to 9200
                ('node03.local', 9201),     # ..or with an explicit port
            ], "admin", "pass")

        Args:
            pool (list of addresses):
                Should be an iterable with node address strings, or tuples
                with `address` and `port` combinations in a tuple or list.
            *auth (str or (str, str)):
                Argument `auth` can be be either a string with a token or a
                tuple with username and password. (the latter may be provided
                as two separate arguments

        Returns:
            asyncio.Future (None):
                Future which should be awaited. The result of the future will
                be set to `None` when successful.

        Remarks:
            Do not use this method if the client is already
            connected. This can be checked with `client.is_connected()`.
        """
        assert self.is_connected() is False
        if len(auth) == 1:
            auth = auth[0]

        self._pool = tuple((
            (address, 9200) if isinstance(address, str) else address
            for address in pool))
        self._auth = self._auth_check(auth)
        self._pool_idx = random.randint(0, len(pool) - 1)
        return self.reconnect()

    def connect(
            self,
            host: str,
            port: int = 9200,
            timeout: Optional[int] = 5
    ) -> asyncio.Future:
        """Connect to ThingsDB.

        This method will *only* create a connection, so the connection is not
        authenticated yet. Use the `authenticate(..)` method after creating a
        connection before using the connection.

        Args:
            host (str):
                A hostname, IP address, FQDN to connect to.
            port (int, optional):
                Integer value between 0 and 65535 and should be the port number
                where a ThingsDB node is listening to for client connections.
                Defaults to 9200.
            timeout (int, optional):
                Can be be used to control the maximum time the client will
                attempt to create a connection. The timeout may be set to
                `None` in which case the client will wait forever on a
                response. Defaults to 5.

        Returns:
            asyncio.Future (None):
                Future which should be awaited. The result of the future will
                be set to `None` when successful.

        Remarks:
            Do not use this method if the client is already
            connected. This can be checked with `client.is_connected()`.
        """
        assert self.is_connected() is False
        self._pool = ((host, port),)
        self._pool_idx = 0
        return self._connect(timeout=timeout)

    async def reconnect(self) -> None:
        """Re-connect to ThingsDB.

        This method can be used, even when a connection still exists. In case
        of a connection pool, a call to `reconnect()` will switch to another
        node.
        """
        if self._reconnecting:
            return

        self._reconnecting = True
        try:
            await self._reconnect_loop()
        finally:
            self._reconnecting = False

    async def wait_closed(self) -> None:
        """Wait for a connection to close.

        Can be used after calling the `close()` method to determine when the
        connection is actually closed.
        """
        if self._protocol and self._protocol.close_future:
            await self._protocol.close_future

    async def authenticate(
            self,
            *auth: Union[str, tuple],
            timeout: Optional[int] = 5
    ) -> None:
        """Authenticate a ThingsDB connection.

        Args:
            *auth (str or (str, str)):
                Argument `auth` can be be either a string with a token or a
                tuple with username and password. (the latter may be provided
                as two separate arguments
            timeout (int, optional):
                Can be be used to control the maximum time in seconds for the
                client to wait for response on the authentication request.
                The timeout may be set to `None` in which case the client will
                wait forever on a response. Defaults to 5.
        """
        if len(auth) == 1:
            auth = auth[0]
        self._auth = self._auth_check(auth)
        await self._authenticate(timeout)

        if self._reconnect:
            try:
                await self.watch(scope='@n')
            except ForbiddenError:
                logging.warning(_WATCH_MISSING)

    def query(
            self,
            code: str,
            scope: Optional[str] = None,
            timeout: Optional[int] = None,
            convert_vars: bool = True,
            **kwargs: Any
    ) -> asyncio.Future:
        """Query ThingsDB.

        Use this method to run `code` in a scope.

        Args:
            code (str):
                ThingsDB code to run.
            scope (str, optional):
                Run the code in this scope. If not specified, the default scope
                will be used. See https://docs.thingsdb.net/v0/overview/scopes/
                for how to format a scope.
            timeout (int, optional):
                Raise a time-out exception if no response is received within X
                seconds. If no time-out is given, the client will wait forever.
                Defaults to `None`.
            convert_vars (bool, optional):
                Only applicable if `**kwargs` are given. If set to `True`, then
                the provided **kwargs values will be converted so ThingsDB can
                understand them. For example, a thing should be given just by
                it's ID and with conversion the `#` will be extracted. When
                this argument is `False`, the **kwargs stay untouched.
                Defaults to `True`.
            **kwargs (any, optional):
                Can be used to inject variable into the ThingsDB code.

        Examples:
            Although we could just as easy have wrote everything in the
            ThingsDB code itself, this example shows how to use **kwargs for
            injecting variable into code. In this case the variable `book`.

            >>> res = await client.query(".my_book = book;", book={
                'title': 'Manual ThingsDB'
            })

        Returns:
            asyncio.Future (any):
                Future which should be awaited. The result of the future will
                contain the result of the ThingsDB code when successful.

        Remarks:
            If the ThingsDB code will return with an exception, then this
            exception will be translated to a Python Exception which will be
            raised. See thingsdb.exceptions for all possible exceptions and
            https://docs.thingsdb.net/v0/errors/ for info on the error codes.
        """
        if self._protocol is None:
            raise ConnectionError('no connection')

        if scope is None:
            scope = self._scope

        if kwargs:
            if convert_vars:
                kwargs = {k: convert(v) for k, v in kwargs.items()}
            data = [scope, code, kwargs]
        else:
            data = [scope, code]

        return self._protocol.write(Proto.REQ_QUERY, data, timeout=timeout)

    def run(
            self,
            procedure: str,
            *args: Optional[Any],
            scope: Optional[str] = None,
            timeout: Optional[int] = None,
            convert_args: bool = True,
            **kwargs: Any,
    ) -> asyncio.Future:
        """Run a procedure.

        Use this method to run a stored procedure in a scope.

        Args:
            procedure (str):
                Name of the procedure to run.
            *args (any):
                Arguments which are injected as the procedure arguments.
                Instead of positional, the arguments may also be parsed using
                keyword arguments but not both at the same time.
            scope (str, optional):
                Run the procedure in this scope. If not specified, the default
                scope will be used.
                See https://docs.thingsdb.net/v0/overview/scopes/ for how to
                format a scope.
            timeout (int, optional):
                Raise a time-out exception if no response is received within X
                seconds. If no time-out is given, the client will wait forever.
                Defaults to `None`.
            convert_args (bool, optional):
                Only applicable if `*args` are given. If set to `True`, then
                the provided `*args` values will be converted so ThingsDB can
                understand them. For example, a thing should be given just by
                it's ID and with conversion the `#` will be extracted. When
                this argument is `False`, the `*args` stay untouched.
                Defaults to `True`.
            **kwargs (any):
                Arguments which are injected as the procedure arguments.
                Instead of by name, the arguments may also be parsed using
                positional arguments but not both at the same time.

        Returns:
            asyncio.Future (any):
                Future which should be awaited. The result of the future will
                contain the result of the ThingsDB procedure when successful.

        Remarks:
            If the ThingsDB code will return with an exception, then this
            exception will be translated to a Python Exception which will be
            raised. See thingsdb.exceptions for all possible exceptions and
            https://docs.thingsdb.net/v0/errors/ for info on the error codes.
        """
        if self._protocol is None:
            raise ConnectionError('no connection')

        if scope is None:
            scope = self._scope

        if kwargs:
            args = {
                k: convert(v)
                for k, v in kwargs.items()
            } if convert_args else kwargs
        elif args and convert_args:
            args = [convert(arg) for arg in args]

        return self._protocol.write(
            Proto.REQ_RUN,
            [scope, procedure, args],
            timeout=timeout)

    def watch(self, *ids: int, scope: Optional[str] = None) -> asyncio.Future:
        """Subscribe for changes on given things.

        This method accepts one or more thing ids to subscribe to. This
        method will simply return None as soon as the subscribe request is
        successful handled by ThingsDB. After the response, the client will
        receive `INIT` events for all subscribed ids. After that, ThingsDB
        will continue to provide the client with `UPDATE` events which contain
        changes to the subscribed thing. A `DELETE` event might be received
        if, and only if the thing is removed and garbage collected from the
        collection.

        Args:
            *ids (int):
                Thing IDs to subscribe to. No error is returned in case one of
                the given things are not found within the collection, instead a
                `WARN` event will be send to the client.
            scope (str, optional):
                Subscribe on things in this scope. If not specified, the
                default scope will be used. Only collection scopes may contain
                things so only collection scopes can be used.
                See https://docs.thingsdb.net/v0/overview/scopes/ for how to
                format a scope.

        Returns:
            asyncio.Future (None):
                Future which result will be set to `None` if successful.
        """
        if self._protocol is None:
            raise ConnectionError('no connection')

        if scope is None:
            scope = self._scope

        return self._protocol.write(Proto.REQ_WATCH, [scope, *ids])

    def unwatch(
            self,
            *ids: int,
            scope: Optional[str] = None
    ) -> asyncio.Future:
        """Unsubscribe for changes on given things.

        Stop receiving events for the things given by one or more ids. It is
        possible that the client receives an event shortly after calling the
        unsubscribe method because the event was queued.

        Args:
            *ids (int):
                Thing IDs to unsubscribe. No error is returned in case one of
                the given things are not found within the collection or if the
                thing was not being watched.
            scope (str, optional):
                Unsubscribe for things in this scope. If not specified, the
                default scope will be used. Only collection scopes may contain
                things so only collection scopes can be used.
                See https://docs.thingsdb.net/v0/overview/scopes/ for how to
                format a scope.

        Returns:
            asyncio.Future (None):
                Future which result will be set to `None` if successful.
        """
        if self._protocol is None:
            raise ConnectionError('no connection')

        if scope is None:
            scope = self._scope

        return self._protocol.write(Proto.REQ_UNWATCH, [scope, *ids])

    @staticmethod
    def _auth_check(auth):
        assert ((
            isinstance(auth, (list, tuple)) and
            len(auth) == 2 and
            isinstance(auth[0], str) and
            isinstance(auth[1], str)
        ) or (
            isinstance(auth, str)
        )), (
            'expecting the authentication argument to be a "token-string" '
            'or a tuple like ("username", "password").'
        )
        return auth

    async def _connect(self, timeout=5):
        host, port = self._pool[self._pool_idx]
        try:
            conn = self._loop.create_connection(
                lambda: Protocol(
                    on_connection_lost=self._on_connection_lost,
                    on_event=self._on_event,
                    loop=self._loop),
                host=host,
                port=port,
                ssl=self._ssl)
            _, self._protocol = await asyncio.wait_for(
                conn,
                timeout=timeout)
        finally:
            self._pool_idx += 1
            self._pool_idx %= len(self._pool)

    def _on_event(self, pkg):
        if pkg.tp == Proto.ON_NODE_STATUS and \
                self._reconnect and \
                pkg.data == 'SHUTTING_DOWN':
            asyncio.ensure_future(self.reconnect(), loop=self._loop)

        for event_handler in self._event_handlers:
            event_handler(pkg.tp, pkg.data)

    def _on_connection_lost(self, protocol, exc):
        if self._protocol is not protocol:
            return

        self._protocol = None

        if self._reconnect:
            asyncio.ensure_future(self.reconnect(), loop=self._loop)

    async def _reconnect_loop(self):
        wait_time = 1
        timeout = 2
        protocol = self._protocol
        while True:
            host, port = self._pool[self._pool_idx]
            try:
                await self._connect(timeout=timeout)
                await self._ping(timeout=2)
            except Exception as e:
                logging.error(
                    f'connecting to {host}:{port} failed ({e}), '
                    f'try next connect in {wait_time} seconds'
                )
            else:
                if protocol and protocol.transport:
                    # make sure the `old` connection will be dropped
                    self._loop.call_later(10.0, protocol.transport.close)
                break

            await asyncio.sleep(wait_time)
            wait_time *= 2
            wait_time = min(wait_time, self.MAX_RECONNECT_WAIT_TIME)
            timeout = min(timeout+1, self.MAX_RECONNECT_TIMEOUT)

        await self._authenticate(timeout=5)

        if self._reconnect:
            try:
                await self.watch(scope='@n')
            except ForbiddenError:
                logging.warning(_WATCH_MISSING)

        for event_handler in self._event_handlers:
            event_handler.on_reconnect()

    def _ping(self, timeout):
        if self._protocol is None:
            raise ConnectionError('no connection')

        return self._protocol.write(Proto.REQ_PING, timeout=timeout)

    def _authenticate(self, timeout):
        if self._protocol is None:
            raise ConnectionError('no connection')

        return self._protocol.write(
            Proto.REQ_AUTH,
            data=self._auth,
            timeout=timeout)

    @deprecated(details='Use `set_default_scope` instead')
    def use(self, scope):
        if not scope.startswith('@'):
            scope = f'@{scope}' if scope.startswith(':') else f'@:{scope}'
        self.set_default_scope(scope)

    @deprecated(details='Use `get_default_scope` instead')
    def get_scope(self):
        return self._scope



import sys
import unittest
sys.path.insert(0, '../')
from thingsdb.util import fmt  # nopep8
from thingsdb.model import Thing, Collection  # nopep8


def fmt_fmt(formatted):
    return f'x={formatted}'


def val_fmt(val, blobs=None):
    if blobs is None:
        blobs = []
    return fmt_fmt(fmt(val, blobs))


class TestWrap(unittest.TestCase):

    def test_fmt_int(self):
        self.assertEqual('x=10', val_fmt(10))
        self.assertEqual('x=-10', val_fmt(-10))

    def test_fmt_float(self):
        self.assertEqual('x=0.5', val_fmt(0.5))
        self.assertEqual('x=-0.5', val_fmt(-0.5))

    def test_fmt_bool(self):
        self.assertEqual('x=true', val_fmt(True))
        self.assertEqual('x=false', val_fmt(False))

    def test_fmt_nil(self):
        self.assertEqual('x=nil', val_fmt(None))

    def test_fmt_string(self):
        self.assertEqual("x='Iris'", val_fmt('Iris'))
        self.assertEqual("x=''", val_fmt(''))

    def test_fmt_list(self):
        self.assertEqual("x=[6,'Iris']", val_fmt([6, 'Iris']))
        self.assertEqual("x=[[6,'Iris']]", val_fmt([[6, 'Iris']]))

    def test_fmt_dict(self):
        self.assertEqual(r"x={}", val_fmt({}))
        self.assertEqual(r"x={name:'Iris'}", val_fmt({'name': 'Iris'}))
        self.assertEqual(r"x={age:6}", val_fmt({'age': 6}))

    def test_fmt_nested(self):
        self.assertEqual(
            r"x=[{stuff:[{lang:'C'},{more:['Python','C',true]}]}]",
            val_fmt([{
                'stuff': [{
                    'lang': 'C',
                }, {
                    'more': ['Python', 'C', True]
                }]
            }])
        )

    def test_blobs(self):
        blobs = {}
        self.assertEqual(r"x=[blob0,blob1]", val_fmt([b'a', b'b'], blobs))
        self.assertEqual(blobs, {'blob0': b'a', 'blob1': b'b'})

    def test_fmt_thing(self):
        t = Thing(Collection(), 42)
        self.assertEqual("x=#42", val_fmt(t))
        self.assertEqual("x=#42", val_fmt({'#': 42, "test": "xyz"}))


if __name__ == '__main__':
    unittest.main()
